dnl
dnl This file is part of the Score-P software (http://www.score-p.org)
dnl
dnl Copyright (c) 2021,
dnl Forschungszentrum Juelich GmbH, Germany
dnl
dnl This software may be modified and distributed under the terms of
dnl a BSD-style license.  See the COPYING file in the package base
dnl directory for details.
dnl


# AFS_SUMMARY_COMPILER_PREFIX(PREFIX)
# -----------------------------------
# Sets the prefix to be used in the configuration summary output of the
# compiler macros `AFS_PROG_CC`, `AFS_PROG_CXX`, `AFS_PROG_F77`, and
# `AFS_PROG_FC` (e.g., "MPI" or "SHMEM").
#
AC_DEFUN([AFS_SUMMARY_COMPILER_PREFIX], [
m4_ifblank([$1], [m4_fatal([$0: empty PREFIX argument])])
_afs_summary_compiler_prefix="$1 "
])



# AFS_PROG_CC([VERSION], [ext|noext], [mandatory|optional])
# ---------------------------------------------------------
# (Arguments follow the `AX_C_COMPILE_STDC` macro, i.e., the documentation
# should be kept in sync!)
#
# Determines a C compiler to use, its vendor and version, and prints this
# information to the configuration summary.
#
# The first argument, if specified, enforces a check for baseline language
# coverage in the compiler for the specified version of the C standard,
# which adds switches to CC and CPP to enable support.  VERSION may be '99'
# (for the C99 standard) or '11' (for the C11 standard).
#
# The second argument is only significant if a specific language VERSION
# is requested.  If specified, indicates whether you insist on an extended
# mode (e.g., '-std=gnu99') or a strict conformance mode (e.g., '-std=c99').
# If neither option is specified, you get whatever works, with preference
# for no added switch, then for an extended mode.
#
# The third argument is also only significant if a specific VERSION is
# requested.  If set to 'mandatory' or left unspecified, indicates that
# baseline support for the given C standard is required and that the macro
# should error out if no mode with that support is found.  If set to
# 'optional', then configuration proceeds regardless, after defining
# 'HAVE_C${VERSION}' if and only if a supporting mode is found.
#
# NOTE:
#   The third argument only affects the C standard conformance test.
#   A working C compiler is always required!
#
# List of configure variables set:
#   `afs_cv_prog_cc_works`::  'yes' if C compiler works, error out otherwise
#   `HAVE_C<VERSION>`::       'yes' if requested language version is supported,
#                             'no' otherwise (only if VERSION is specified)
#
AC_DEFUN([AFS_PROG_CC], [
AC_REQUIRE([AC_PROG_CC])
AC_BEFORE([$0], [AFS_PROG_CXX])
AC_BEFORE([$0], [AFS_PROG_F77])
AC_BEFORE([$0], [AFS_PROG_FC])

dnl If we get here, we know that the C compiler is working
AC_CACHE_VAL([afs_cv_prog_cc_works], [afs_cv_prog_cc_works=yes])
AC_LANG_PUSH([C])

dnl Feature checks
m4_ifnblank([$1], [AX_C_COMPILE_STDC([$1], [$2], [$3])])

dnl Vendor & version
AX_COMPILER_VENDOR
AX_COMPILER_VERSION

dnl Compiler-specific quirks
dnl
dnl The "classic" Cray C compilers prior to CCE 9.0 enable OpenMP by default.
dnl Thus enforce non-OpenMP mode to allow for explicit control, which doesn't
dnl do any harm for later versions.
AS_IF([test x${ax_cv_c_compiler_vendor%/*} = xcray],
      [CC="$CC -hnoomp"])

AC_LANG_POP([C])

dnl Print summary output
_afs_prog_cc_info=" ($ax_cv_c_compiler_vendor $ax_cv_c_compiler_version)"
AFS_SUMMARY([${_afs_summary_compiler_prefix}C$1 compiler],
            [$CC$_afs_prog_cc_info])
AS_UNSET([_afs_prog_cc_info])

dnl All compilers we care about support '-c -o', i.e., use of the 'compile'
dnl wrapper script is not necessary.  However, 'autoreconf' is complaining
dnl if the 'AM_PROG_CC_C_O' macro is not used.  It is intentionally placed
dnl after the summary output to not mess it up with the 'compile' script
dnl prefix.
dnl
dnl In case we encounter a compiler that does not support '-c -o' in the
dnl future, similar modifications to what this macro does are required for
dnl '$CXX', '$F77', and '$FC'.
AM_PROG_CC_C_O
])


# AFS_PROG_CXX([VERSION], [ext|noext], [mandatory|optional])
# ----------------------------------------------------------
# (Arguments follow the `AX_CXX_COMPILE_STDCXX` macro, i.e., the documentation
# should be kept in sync!)
#
# Determines a C++ compiler to use, its vendor and version, and prints this
# information to the configuration summary (if applicable).
#
# The first argument, if specified, enables a check for baseline language
# coverage in the compiler for the specified version of the C++ standard,
# which adds switches to CXX and CXXCPP to enable support.  VERSION may be
# '11' (for the C++11 standard), '14' (for the C++14 standard), or '17'
# (for the C++17 standard).
#
# The second argument is only significant if a specific language VERSION
# is requested.  If specified, indicates whether you insist on an extended
# mode (e.g., '-std=gnu++11') or a strict conformance mode (e.g.,
# '-std=c++11').  If neither option is specified, you get whatever works,
# with preference for no added switch, then for an extended mode.
#
# The third argument, if set to 'mandatory' or left unspecified, indicates
# that a working C++ compiler with baseline support for given language
# VERSION is required, and that the macro should error out if either no
# working compiler or no compiler with that support is found.  If set to
# 'optional', then configuration proceeds regardless, after defining
# 'HAVE_CXX${VERSION}' if and only if a working compiler with supporting
# mode is found.
#
# List of configure variables set:
#   `afs_cv_prog_cxx_works`::  'yes' if C++ compiler works, 'no' otherwise
#   `HAVE_CXX<VERSION>`::      'yes' if requested language version is supported,
#                              'no' otherwise (only if VERSION is specified)
#
AC_DEFUN([AFS_PROG_CXX], [
AC_REQUIRE([AC_PROG_CXX])

m4_if([$3], [], [afs_prog_cxx_required=true],
      [$3], [mandatory], [afs_prog_cxx_required=true],
      [$3], [optional], [afs_prog_cxx_required=false],
      [m4_fatal([invalid third argument `$3' to AFS_PROG_CXX])])
AS_IF([test -z "$CXX"], [
    dnl Error out if C++ compiler is mandatory but could not be found
    AS_IF([test x$afs_prog_cxx_required = xtrue],
          [AC_MSG_FAILURE([no acceptable C++ compiler found in \$PATH])])
    CXX=no
    AC_CACHE_VAL([afs_cv_prog_cxx_works], [afs_cv_prog_cxx_works=no])
],[
    AC_LANG_PUSH([C++])

    dnl Check for working compiler
    AC_CACHE_CHECK([whether the C++ compiler works],
        [afs_cv_prog_cxx_works],
        [AC_LINK_IFELSE([AC_LANG_PROGRAM([], [])],
                        [afs_cv_prog_cxx_works=yes],
                        [afs_cv_prog_cxx_works=no])])
    AS_IF([test x$afs_cv_prog_cxx_works = xno], [
        dnl Error out if C++ compiler is mandatory but does not work
        AS_IF([test x$afs_prog_cxx_required = xtrue],
              [AC_MSG_FAILURE([C++ compiler cannot create executables])])
        CXX=no
    ],[
        dnl Feature checks
        m4_ifnblank([$1], [AX_CXX_COMPILE_STDCXX([$1], [$2], [$3])])
        AX_CXX_INTTYPE_MACROS

        dnl Vendor & version
        AX_COMPILER_VENDOR
        AX_COMPILER_VERSION

        dnl Compiler-specific quirks
        dnl
        dnl The "classic" Cray C++ compilers prior to CCE 9.0 enable OpenMP
        dnl by default.  Thus enforce non-OpenMP mode to allow for explicit
        dnl control, which doesn't do any harm for later versions.
        AS_IF([test x${ax_cv_cxx_compiler_vendor%/*} = xcray],
              [CXX="$CXX -hnoomp"])
    ])

    AC_LANG_POP([C++])
])

dnl Print summary output
AS_UNSET([_afs_prog_cxx_info])
AS_IF([test x$afs_cv_prog_cxx_works = xyes],
      [_afs_prog_cxx_info=" ($ax_cv_cxx_compiler_vendor $ax_cv_cxx_compiler_version)"])
AFS_SUMMARY([${_afs_summary_compiler_prefix}C++$1 compiler],
            [$CXX$_afs_prog_cxx_info])
AS_UNSET([_afs_prog_cxx_info])
])


# AFS_PROG_F77([mandatory|optional])
# ----------------------------------
# Determines a Fortran 77 compiler to use and prints this information to the
# configuration summary (if applicable).
#
# The first argument, if set to 'mandatory' or left unspecified, indicates
# that a working Fortran 77 compiler is required, and that the macro should
# error out if no compiler is found.  If set to 'optional', then configuration
# proceeds regardless.
#
# List of configure variables set:
#   `afs_cv_prog_f77_works`::  'yes' if Fortran 77 compiler works, 'no'
#                              otherwise
#
AC_DEFUN([AFS_PROG_F77], [
AC_REQUIRE([AC_PROG_F77])

m4_if([$1], [], [afs_prog_f77_required=true],
      [$1], [mandatory], [afs_prog_f77_required=true],
      [$1], [optional], [afs_prog_f77_required=false],
      [m4_fatal([invalid first argument `$1' to AFS_PROG_F77])])
AS_IF([test -z "$F77"], [
    dnl Error out if Fortran 77 compiler is mandatory but could not be found
    AS_IF([test x$afs_prog_f77_required = xtrue],
          [AC_MSG_FAILURE([no acceptable Fortran 77 compiler found in \$PATH])])
    F77=no
    AC_CACHE_VAL([afs_cv_prog_f77_works], [afs_cv_prog_f77_works=no])
],[
    AC_LANG_PUSH([Fortran 77])

    dnl Check for working compiler
    AC_CACHE_CHECK([whether the Fortran 77 compiler works],
        [afs_cv_prog_f77_works],
        [AC_LINK_IFELSE([AC_LANG_PROGRAM([], [])],
                        [afs_cv_prog_f77_works=yes],
                        [afs_cv_prog_f77_works=no])])
    AS_IF([test x$afs_cv_prog_f77_works = xno], [
        dnl Error out if Fortran 77 compiler is mandatory but does not work
        AS_IF([test x$afs_prog_f77_required = xtrue],
              [AC_MSG_FAILURE([Fortran 77 compiler cannot create executables])])
        F77=no
    ],[
        dnl Compiler-specific quirks
        dnl
        dnl The Cray Fortran compilers prior to CCE 9.0 enable OpenMP by
        dnl default.  Thus enforce non-OpenMP mode to allow for explicit
        dnl control, which doesn't do any harm for later versions.
        dnl
        dnl NOTE: This decision is based on the C compiler vendor, as the
        dnl       'AX_COMPILER_VENDOR' macro only works for $FC !
        AS_IF([test x${ax_cv_c_compiler_vendor%/*} = xcray],
              [F77="$F77 -hnoomp"])
    ])

    AC_LANG_POP([Fortran 77])
])

dnl Print summary output
AFS_SUMMARY([${_afs_summary_compiler_prefix}F77 compiler], [$F77])
])


# AFS_PROG_FC([mandatory|optional])
# ---------------------------------
# Determines a Fortran compiler to use, its vendor, and prints this information
# to the configuration summary (if applicable).
#
# The first argument, if set to 'mandatory' or left unspecified, indicates
# that a working Fortran compiler is required, and that the macro should
# error out if no compiler is found.  If set to 'optional', then configuration
# proceeds regardless.
#
# List of configure variables set:
#   `afs_cv_prog_fc_works`::  'yes' if Fortran compiler works, 'no' otherwise
#
AC_DEFUN([AFS_PROG_FC], [
AC_REQUIRE([AC_PROG_FC])

m4_if([$1], [], [afs_prog_fc_required=true],
      [$1], [mandatory], [afs_prog_fc_required=true],
      [$1], [optional], [afs_prog_fc_required=false],
      [m4_fatal([invalid first argument `$1' to AFS_PROG_FC])])
AS_IF([test -z "$FC"], [
    dnl Error out if Fortran compiler is mandatory but could not be found
    AS_IF([test x$afs_prog_fc_required = xtrue],
          [AC_MSG_FAILURE([no acceptable Fortran compiler found in \$PATH])])
    FC=no
    AC_CACHE_VAL([afs_cv_prog_fc_works], [afs_cv_prog_fc_works=no])
],[
    AC_LANG_PUSH([Fortran])

    dnl Check for working compiler
    AC_CACHE_CHECK([whether the Fortran compiler works],
        [afs_cv_prog_fc_works],
        [AC_LINK_IFELSE([AC_LANG_PROGRAM([], [])],
                        [afs_cv_prog_fc_works=yes],
                        [afs_cv_prog_fc_works=no])])
    AS_IF([test x$afs_cv_prog_fc_works = xno], [
        dnl Error out if Fortran compiler is mandatory but does not work
        AS_IF([test x$afs_prog_fc_required = xtrue],
              [AC_MSG_FAILURE([Fortran compiler cannot create executables])])
        FC=no
    ],[
        dnl Feature checks
        AC_FC_PP_SRCEXT([F])

        dnl Vendor
        AX_COMPILER_VENDOR

        dnl Compiler-specific quirks
        dnl
        dnl The Cray Fortran compilers prior to CCE 9.0 enable OpenMP by
        dnl default.  Thus enforce non-OpenMP mode to allow for explicit
        dnl control, which doesn't do any harm for later versions.
        AS_IF([test x${ax_cv_fc_compiler_vendor%/*} = xcray],
              [FC="$FC -hnoomp"])
    ])

    AC_LANG_POP([Fortran])
])

dnl Print summary output
AS_UNSET([_afs_prog_fc_info])
AS_IF([test x$afs_cv_prog_fc_works = xyes],
      [_afs_prog_fc_info=" ($ax_cv_fc_compiler_vendor)"])
AFS_SUMMARY([${_afs_summary_compiler_prefix}Fortran compiler],
            [$FC$_afs_prog_fc_info])
AS_UNSET([_afs_prog_fc_info])
])


## TODO: REMOVE THIS MOCKUP!
AC_DEFUN([AFS_SUMMARY], [
    AC_MSG_NOTICE(["$1: $2"])
])
