/*
 * Copyright (c) 2022,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>
#include "ompt_tool.h"

#include <omp-tools.h>
#include <stdlib.h>
#include <assert.h>
#include <inttypes.h>
#include <string.h>

_Thread_local int32_t ompt_tool_tid          = -1; /* thread counter */
atomic_flag           ompt_tool_printf_mutex = ATOMIC_FLAG_INIT;


/*
   static const char*
   callbacks2string( ompt_callbacks_t t )
   {
    switch ( t )
    {
   #if 0
   ompt_callback_thread_begin            = 1,
   ompt_callback_thread_end              = 2,
   ompt_callback_parallel_begin          = 3,
   ompt_callback_parallel_end            = 4,
   ompt_callback_task_create             = 5,
   ompt_callback_task_schedule           = 6,
   ompt_callback_implicit_task           = 7,
   ompt_callback_target                  = 8,
   ompt_callback_target_data_op          = 9,
   ompt_callback_target_submit           = 10,
   ompt_callback_control_tool            = 11,
   ompt_callback_device_initialize       = 12,
   ompt_callback_device_finalize         = 13,
   ompt_callback_device_load             = 14,
   ompt_callback_device_unload           = 15,
   ompt_callback_sync_region_wait        = 16,
   ompt_callback_mutex_released          = 17,
   ompt_callback_dependences             = 18,
   ompt_callback_task_dependence         = 19,
   ompt_callback_work                    = 20,
   ompt_callback_masked                  = 21,
   ompt_callback_master                   = ompt_callback_masked, // deprecated
   ompt_callback_target_map              = 22,
   ompt_callback_sync_region             = 23,
   ompt_callback_lock_init               = 24,
   ompt_callback_lock_destroy            = 25,
   ompt_callback_mutex_acquire           = 26,
   ompt_callback_mutex_acquired          = 27,
   ompt_callback_nest_lock               = 28,
   ompt_callback_flush                   = 29,
   ompt_callback_cancel                  = 30,
   ompt_callback_reduction               = 31,
   ompt_callback_dispatch                = 32,
   ompt_callback_target_emi              = 33,
   ompt_callback_target_data_op_emi      = 34,
   ompt_callback_target_submit_emi       = 35,
   ompt_callback_target_map_emi          = 36,
   ompt_callback_error                   = 37
   #endif
    }
    assert( false );
    return "";
   }
 */

static const char*
cancel_flag2string( int /* ompt_cancel_flag_t */ t )
{
    /*
       ompt_cancel_parallel = 0x01,
       ompt_cancel_sections = 0x02,
       ompt_cancel_loop = 0x04,
       ompt_cancel_taskgroup = 0x08,
       ompt_cancel_activated = 0x10,
       ompt_cancel_detected = 0x20,
       ompt_cancel_discarded_task = 0x40
     */

    /* The flags argument, defined by the ompt_cancel_flag_t
       enumeration type, indicates whether cancellation is activated
       by the current task or detected as being activated by another
       task. The construct that is being canceled is also described in
       the flags argument. When several constructs are detected as
       being concurrently canceled, each corresponding bit in the
       argument will be set. */

    #define PARALLEL "_parallel"
    #define SECTIONS "_sections"
    #define LOOP "_loop"
    #define TASKGROUP "_taskgroup"

    #define ACTIVATED "activated"
    #define DETECTED  "detected"
    #define DISCARDED "discarded"

    assert( ( sizeof( DISCARDED ) >= sizeof( DETECTED ) ) &&
            ( sizeof( DISCARDED ) >= sizeof( ACTIVATED ) ) );
    static __thread char str[ sizeof( DISCARDED ) + sizeof( PARALLEL ) + sizeof( SECTIONS ) + sizeof( LOOP ) + sizeof( TASKGROUP ) + 1 ];
    size_t               offset;
    if ( t & ompt_cancel_activated )
    {
        memcpy( str, ACTIVATED, sizeof( ACTIVATED ) );
        offset = sizeof( ACTIVATED );
    }
    else if ( t & ompt_cancel_detected )
    {
        memcpy( str, DETECTED, sizeof( DETECTED ) );
        offset = sizeof( DETECTED );
    }
    else if ( t & ompt_cancel_discarded_task )
    {
        memcpy( str, DISCARDED, sizeof( DISCARDED ) );
        offset = sizeof( DISCARDED );
    }
    else
    {
        assert( false );
    }

    bool cancelled_where = false;
    if ( t & ompt_cancel_parallel )
    {
        memcpy( str + offset, PARALLEL, sizeof( PARALLEL ) );
        offset         += sizeof( PARALLEL );
        cancelled_where = true;
    }
    if ( t & ompt_cancel_sections )
    {
        memcpy( str + offset, SECTIONS, sizeof( SECTIONS ) );
        offset         += sizeof( SECTIONS );
        cancelled_where = true;
    }
    if ( t & ompt_cancel_loop )
    {
        memcpy( str + offset, LOOP, sizeof( LOOP ) );
        offset         += sizeof( LOOP );
        cancelled_where = true;
    }
    if ( t & ompt_cancel_discarded_task )
    {
        memcpy( str + offset, DISCARDED, sizeof( DISCARDED ) );
        offset         += sizeof( DISCARDED );
        cancelled_where = true;
    }
    assert( cancelled_where );
    str[ offset ] = '\0';
    return str;

    #undef PARALLEL
    #undef SECTIONS
    #undef LOOP
    #undef TASKGROUP

    #undef ACTIVATED
    #undef DETECTED
    #undef DISCARDED
}

static const char*
dependence_type2string( ompt_dependence_type_t t )
{
    switch ( t )
    {
        case ompt_dependence_type_in:
            return "in";
        case ompt_dependence_type_out:
            return "out";
        case ompt_dependence_type_inout:
            return "inout";
        case ompt_dependence_type_mutexinoutset:
            return "mutexinoutset";
        case ompt_dependence_type_source:
            return "source";
        case ompt_dependence_type_sink:
            return "sink";
        #if HAVE_DECL_OMPT_DEPENDENCE_TYPE_INOUTSET
        case ompt_dependence_type_inoutset:
            return "inoutset";
        #endif
    }
    assert( false );
    return "";
}

static const char*
dispatch2string( ompt_dispatch_t t )
{
    switch ( t )
    {
        case ompt_dispatch_iteration:
            return "iteration";
        case ompt_dispatch_section:
            return "section";
        #if HAVE_DECL_OMPT_DISPATCH_WS_LOOP_CHUNK
        case ompt_dispatch_ws_loop_chunk:
            return "ws_loop_chunk";
        #endif
        #if HAVE_DECL_OMPT_DISPATCH_TASKLOOP_CHUNK
        case ompt_dispatch_taskloop_chunk:
            return "taskloop_chunk";
        #endif
        #if HAVE_DECL_OMPT_DISPATCH_DISTRIBUTE_CHUNK
        case ompt_dispatch_distribute_chunk:
            return "distribute_chunk";
        #endif
    }
    assert( false );
    return "";
}

static const char*
frame_flag2string( ompt_frame_flag_t t )
{
    /*
       ompt_frame_runtime = 0x00,
       ompt_frame_application = 0x01,
       ompt_frame_cfa = 0x10,
       ompt_frame_framepointer = 0x20,
       ompt_frame_stackaddress = 0x30
     */

    #define FRAME_FLAGS \
    FRAME_FLAG( runtime ) \
    FRAME_FLAG( application ) \
    FRAME_FLAG( cfa ) \
    FRAME_FLAG( framepointer ) \
    FRAME_FLAG( stackaddress )

    #define FRAME_FLAG( NAME )  sizeof( #NAME ) +
    static __thread char str[ FRAME_FLAGS + 1 ];
    #undef FRAME_FLAG

    size_t offset = 0;

    #define FRAME_FLAG( NAME )               \
    if ( t & ompt_frame_ ## NAME ) \
    { \
        if ( offset ) \
        { \
            memcpy( str, "_" #NAME, 1 + sizeof( #NAME ) ); \
            offset += 1 + sizeof( #NAME ); \
        } \
        else \
        { \
            memcpy( str, #NAME, sizeof( #NAME ) ); \
            offset += sizeof( #NAME ); \
        } \
    }
    FRAME_FLAGS
    #undef FRAME_FLAG

    assert( offset );

    str[ offset ] = '\0';
    return str;

    #undef FRAME_FLAGS
}

static const char*
mutex2string( ompt_mutex_t t )
{
    switch ( t )
    {
        case ompt_mutex_lock:
            return "lock";
        case ompt_mutex_test_lock:
            return "test_lock";
        case ompt_mutex_nest_lock:
            return "nest_lock";
        case ompt_mutex_test_nest_lock:
            return "test_nest_lock";
        case ompt_mutex_critical:
            return "critical";
        case ompt_mutex_atomic:
            return "atomic";
        case ompt_mutex_ordered:
            return "ordered";
    }
    assert( false );
    return "";
}

static const char*
native_mon_flag2string( ompt_native_mon_flag_t t )
{
    /*
       ompt_native_data_motion_explicit = 0x01,
       ompt_native_data_motion_implicit = 0x02,
       ompt_native_kernel_invocation = 0x04,
       ompt_native_kernel_execution = 0x08,
       ompt_native_driver = 0x10,
       ompt_native_runtime = 0x20,
       ompt_native_overhead = 0x40,
       ompt_native_idleness = 0x80
     */

    #define NATIVE_MON_FLAGS \
    NATIVE_MON_FLAG( data_motion_explicit ) \
    NATIVE_MON_FLAG( data_motion_implicit ) \
    NATIVE_MON_FLAG( kernel_invocation ) \
    NATIVE_MON_FLAG( kernel_execution ) \
    NATIVE_MON_FLAG( driver ) \
    NATIVE_MON_FLAG( runtime ) \
    NATIVE_MON_FLAG( overhead ) \
    NATIVE_MON_FLAG( idleness )

    #define NATIVE_MON_FLAG( NAME )  sizeof( #NAME ) +
    static __thread char str[ NATIVE_MON_FLAGS + 1 ];
    #undef NATIVE_MON_FLAG

    size_t offset = 0;

    #define NATIVE_MON_FLAG( NAME )                  \
    if ( t & ompt_native_ ## NAME ) \
    { \
        if ( offset ) \
        { \
            memcpy( str, "_" #NAME, 1 + sizeof( #NAME ) ); \
            offset += 1 + sizeof( #NAME ); \
        } \
        else \
        { \
            memcpy( str, #NAME, sizeof( #NAME ) ); \
            offset += sizeof( #NAME ); \
        } \
    }
    NATIVE_MON_FLAGS
    #undef NATIVE_MON_FLAG

    assert( offset );

    str[ offset ] = '\0';
    return str;

    #undef NATIVE_MON_FLAGS
}

static const char*
parallel_flag2string( int /* ompt_parallel_flag_t */ t )
{
    /*
       The value ompt_parallel_invoker_program indicates that the
       outlined function associated with implicit tasks for the region
       is invoked directly by the application on the primary thread
       for a parallel region. The value ompt_parallel_invoker_runtime
       indicates that the outlined function associated with implicit
       tasks for the region is invoked by the runtime on the primary
       thread for a parallel region.

       The value ompt_parallel_league indicates that the callback is
       invoked due to the creation of a league of teams by a teams
       construct. The value ompt_parallel_team indicates that the
       callback is invoked due to the creation of a team of threads by
       a parallel construct.
     */
    if ( t & ompt_parallel_invoker_program )
    {
        if ( t & ompt_parallel_league )
        {
            assert( t == ( ompt_parallel_invoker_program | ompt_parallel_league ) );
            return "invoker_program_league";
        }
        else if ( t & ompt_parallel_team )
        {
            assert( t == ( ompt_parallel_invoker_program | ompt_parallel_team ) );
            return "invoker_program_team";
        }
        else
        {
            assert( false );
        }
    }
    else if ( t & ompt_parallel_invoker_runtime )
    {
        if ( t & ompt_parallel_league )
        {
            assert( t == ( ompt_parallel_invoker_runtime | ompt_parallel_league ) );
            return "invoker_runtime_league";
        }
        else if ( t & ompt_parallel_team )
        {
            assert( t == ( ompt_parallel_invoker_runtime | ompt_parallel_team ) );
            return "invoker_runtime_team";
        }
        else
        {
            assert( false );
        }
    }
    else
    {
        assert( false );
    }
    return "";
#if 0
    #define IPRG "invoker_program_"
    #define IRT "invoker_runtime_"
    #define LEAGUE "league"
    #define TEAM "team"

    assert( strlen( IPRG ) == strlen( IRT ) );
    static __thread char str[ strlen( IPRG ) + strlen( LEAGUE ) + 1 ];

    bool invoker = false;
    if ( t & ompt_parallel_invoker_program )
    {
        memcpy( str, IPRG, strlen( IPRG ) );
        invoker = true;
    }
    else if ( t & ompt_parallel_invoker_runtime )
    {
        memcpy( str, IRT, strlen( IRT ) );
        invoker = true;
    }
    assert( invoker );

    bool kind = false;
    if ( t & ompt_parallel_league )
    {
        memcpy( str + strlen( IPRG ), LEAGUE, strlen( LEAGUE ) );
        str[ strlen( IPRG ) + strlen( LEAGUE ) ] = '\0';
        kind                                     = true;
    }
    else if ( t & ompt_parallel_team )
    {
        memcpy( str + strlen( IPRG ), TEAM, strlen( TEAM ) );
        str[ strlen( IPRG ) + strlen( TEAM ) ] = '\0';
        kind                                   = true;
    }
    assert( kind );
    return str;

    #undef IPRG
    #undef IRT
    #undef LEAGUE
    #undef TEAM
#endif
}

static const char*
record_native2string( ompt_record_native_t t )
{
    switch ( t )
    {
        case ompt_record_native_info:
            return "info";
        case ompt_record_native_event:
            return "event";
    }
    assert( false );
    return "";
}

static const char*
record2string( ompt_record_t t )
{
    switch ( t )
    {
        case ompt_record_ompt:
            return "ompt";
        case ompt_record_native:
            return "native";
        case ompt_record_invalid:
            return "invalid";
    }
    assert( false );
    return "";
}

static const char*
scope_endpoint2string( ompt_scope_endpoint_t t )
{
    switch ( t )
    {
        case ompt_scope_begin:
            return "begin";
        case ompt_scope_end:
            return "end";
        #if HAVE_DECL_OMPT_SCOPE_BEGINEND
        case ompt_scope_beginend:
            return "beginend";
        #endif
    }
    assert( false );
    return "";
}

static const char*
set_result2string( ompt_set_result_t t )
{
    switch ( t )
    {
        case ompt_set_error:
            return "error";
        case ompt_set_never:
            return "never";
        case ompt_set_impossible:
            return "impossible";
        case ompt_set_sometimes:
            return "sometimes";
        case ompt_set_sometimes_paired:
            return "sometimes_paired";
        case ompt_set_always:
            return "always";
    }
    assert( false );
    return "";
}

#if HAVE_OMPT_SEVERITY_T
static const char*
severity2string( ompt_severity_t t )
{
    switch ( t )
    {
        case ompt_warning:
            return "warning";
        case ompt_fatal:
            return "fatal";
    }
    assert( false );
    return "";
}
#endif

static const char*
state2string( ompt_state_t t )
{
    /*
       ompt_state_work_serial = 0x000,
       ompt_state_work_parallel = 0x001,
       ompt_state_work_reduction = 0x002,

       ompt_state_wait_barrier = 0x010, // deprecated
       ompt_state_wait_barrier_implicit_parallel = 0x011,
       ompt_state_wait_barrier_implicit_workshare = 0x012,
       ompt_state_wait_barrier_implicit = 0x013, // deprecated
       ompt_state_wait_barrier_explicit = 0x014,
       ompt_state_wait_barrier_implementation = 0x015,
       ompt_state_wait_barrier_teams = 0x016,

       ompt_state_wait_taskwait = 0x020,
       ompt_state_wait_taskgroup = 0x021,

       ompt_state_wait_mutex = 0x040,
       ompt_state_wait_lock = 0x041,
       ompt_state_wait_critical = 0x042,
       ompt_state_wait_atomic = 0x043,
       ompt_state_wait_ordered = 0x044,

       ompt_state_wait_target = 0x080,
       ompt_state_wait_target_map = 0x081,
       ompt_state_wait_target_update = 0x082,

       ompt_state_idle = 0x100,
       ompt_state_overhead = 0x101,
       ompt_state_undefined = 0x102
     */
    #define STATES \
    STATE( work_serial ) \
    STATE( work_parallel ) \
    STATE( work_reduction ) \
    STATE( wait_barrier_implicit_parallel ) \
    STATE( wait_barrier_implicit_workshare ) \
    STATE( wait_barrier_explicit ) \
    STATE( wait_barrier_implementation ) \
    STATE( wait_barrier_teams ) \
    STATE( wait_taskwait ) \
    STATE( wait_taskgroup ) \
    STATE( wait_mutex ) \
    STATE( wait_lock ) \
    STATE( wait_critical ) \
    STATE( wait_atomic ) \
    STATE( wait_ordered ) \
    STATE( wait_target ) \
    STATE( wait_target_map ) \
    STATE( wait_target_update ) \
    STATE( idle ) \
    STATE( overhead ) \
    STATE( undefined )

    #define STATE( NAME )  sizeof( #NAME ) +
    static __thread char str[ STATES 1 ];
    #undef STATE

    size_t offset = 0;

    #define STATE( NAME )                    \
    if ( t & ompt_state_ ## NAME ) \
    { \
        if ( offset ) \
        { \
            memcpy( str, "_" #NAME, 1 + sizeof( #NAME ) ); \
            offset += 1 + sizeof( #NAME ); \
        } \
        else \
        { \
            memcpy( str, #NAME, sizeof( #NAME ) ); \
            offset += sizeof( #NAME ); \
        } \
    }
    STATES
    #undef STATE

    assert( offset );

    str[ offset ] = '\0';
    return str;

    #undef STATES
}

static const char*
sync_region2string( ompt_sync_region_t t )
{
    switch ( t )
    {
        case ompt_sync_region_barrier:
            return "barrier (deprecated)";
        case ompt_sync_region_barrier_implicit:
            return "barrier_implicit (deprecated)";
        case ompt_sync_region_barrier_explicit:
            return "barrier_explicit";
        case ompt_sync_region_barrier_implementation:
            return "barrier_implementation";
        case ompt_sync_region_taskwait:
            return "taskwait";
        case ompt_sync_region_taskgroup:
            return "taskgroup";
        case ompt_sync_region_reduction:
            return "reduction";
        #if HAVE_DECL_OMPT_SYNC_REGION_BARRIER_IMPLICIT_WORKSHARE
        case ompt_sync_region_barrier_implicit_workshare:
            return "barrier_implicit_workshare";
        #endif
        #if HAVE_DECL_OMPT_SYNC_REGION_BARRIER_IMPLICIT_PARALLEL
        case ompt_sync_region_barrier_implicit_parallel:
            return "barrier_implicit_parallel";
        #endif
        #if HAVE_DECL_OMPT_SYNC_REGION_BARRIER_TEAMS
        case ompt_sync_region_barrier_teams:
            return "barrier_teams";
        #endif
    }
    assert( false );
    return "";
}

static const char*
target_data_op2string( ompt_target_data_op_t t )
{
    switch ( t )
    {
        case ompt_target_data_alloc:
            return "alloc";
        case ompt_target_data_transfer_to_device:
            return "transfer_to_device";
        case ompt_target_data_transfer_from_device:
            return "transfer_from_device";
        case ompt_target_data_delete:
            return "delete";
        case ompt_target_data_associate:
            return "associate";
        case ompt_target_data_disassociate:
            return "disassociate";
        #if HAVE_DECL_OMPT_TARGET_DATA_ALLOC_ASYNC
        case ompt_target_data_alloc_async:
            return "alloc_async";
        #endif
        #if HAVE_DECL_OMPT_TARGET_DATA_TRANSFER_TO_DEVICE_ASYNC
        case ompt_target_data_transfer_to_device_async:
            return "transfer_to_device_async";
        #endif
        case ompt_target_data_transfer_from_device_async:
            return "transfer_from_device_async";
        case ompt_target_data_delete_async:
            return "delete_async";
    }
    assert( false );
    return "";
}

static const char*
target_map_flag2string( unsigned int /* ompt_target_map_flag_t */ t )
{
    /*
       ompt_target_map_flag_to = 0x01,
       ompt_target_map_flag_from = 0x02,
       ompt_target_map_flag_alloc = 0x04,
       ompt_target_map_flag_release = 0x08,
       ompt_target_map_flag_delete = 0x10,
       ompt_target_map_flag_implicit = 0x20,
       // since 5.2:
       ompt_target_map_flag_always = 0x40,
       ompt_target_map_flag_present = 0x80,
       ompt_target_map_flag_close = 0x100,
       ompt_target_map_flag_shared = 0x200
     */

    #define MAP_FLAGS \
    MAP_FLAG( to ) \
    MAP_FLAG( from ) \
    MAP_FLAG( alloc ) \
    MAP_FLAG( release ) \
    MAP_FLAG( delete ) \
    MAP_FLAG( implicit )

    /* added in 5.2:
        MAP_FLAG( always ) \
        MAP_FLAG( present ) \
        MAP_FLAG( close ) \
        MAP_FLAG(shared )
     */
    assert( t <= ompt_target_map_flag_implicit ); /* activate enum values added in 5.2 */

    #define MAP_FLAG( NAME )  sizeof( #NAME ) +
    static __thread char str[ MAP_FLAGS + 1 ];
    #undef MAP_FLAG

    size_t offset = 0;

    #define MAP_FLAG( NAME ) \
    if ( t & ompt_target_map_flag_ ## NAME ) \
    { \
        if ( offset ) \
        { \
            memcpy( str, "_" #NAME, 1 + sizeof( #NAME ) ); \
            offset += 1 + sizeof( #NAME ); \
        } \
        else \
        { \
            memcpy( str, #NAME, sizeof( #NAME ) ); \
            offset += sizeof( #NAME ); \
        } \
    }
    MAP_FLAGS
    #undef MAP_FLAG

    assert( offset );

    str[ offset ] = '\0';
    return str;

    #undef MAP_FLAGS
}

static const char*
target2string( ompt_target_t t )
{
    switch ( t )
    {
        case ompt_target:
            return "target";
        case ompt_target_enter_data:
            return "target_enter_data";
        case ompt_target_exit_data:
            return "target_exit_data";
        case ompt_target_update:
            return "target_update";
        #if HAVE_DECL_OMPT_TARGET_NOWAIT
        case ompt_target_nowait:
            return "target_nowait";
        #endif
        #if HAVE_DECL_OMPT_TARGET_NOWAIT
        case ompt_target_enter_data_nowait:
            return "target_enter_data_nowait";
        #endif
        #if HAVE_DECL_OMPT_TARGET_EXIT_DATA_NOWAIT
        case ompt_target_exit_data_nowait:
            return "target_exit_data_nowait";
        #endif
        #if HAVE_DECL_OMPT_TARGET_UPDATE_NOWAIT
        case ompt_target_update_nowait:
            return "target_update_nowait";
        #endif
    }
    assert( false );
    return "";
}

static const char*
task_flag2string( int /* ompt_task_flag_t */ t )
{
    /*
       ompt_task_initial = 0x00000001,
       ompt_task_implicit = 0x00000002,
       ompt_task_explicit = 0x00000004,
       ompt_task_target = 0x00000008,
       ompt_task_taskwait = 0x00000010,
       ompt_task_undeferred = 0x08000000,
       ompt_task_untied = 0x10000000,
       ompt_task_final = 0x20000000,
       ompt_task_mergeable = 0x40000000,
       ompt_task_merged = 0x80000000
     */

    /* The flags argument indicates the kind of task (explicit or
       target) that is generated. Values for flags are a disjunction
       of elements in the ompt_task_flag_t enumeration type.

       The ompt_task_flag_t enumeration type defines valid task type
       values. The least significant byte provides information about
       the general classification of the task. The other bits
       represent properties of the task.
     */
    if ( t & ompt_task_initial )
    {
        assert( t == ompt_task_initial );
        return "initial";
    }
    else if ( t & ompt_task_implicit )
    {
        if ( t & ompt_task_undeferred )
        {
            return "implicit_undeferred";
        }
        else if ( t & ompt_task_untied )
        {
            return "implicit_untied";
        }
        else if ( t & ompt_task_final )
        {
            return "implicit_final";
        }
        else if ( t & ompt_task_mergeable )
        {
            return "implicit_mergeable";
        }
        else if ( t & ompt_task_merged )
        {
            return "implicit_merged";
        }
        else
        {
            assert( t == ompt_task_implicit );
            return "implicit";
        }
    }
    else if ( t & ompt_task_explicit )
    {
        if ( t & ompt_task_undeferred )
        {
            return "explicit_undeferred";
        }
        else if ( t & ompt_task_untied )
        {
            return "explicit_untied";
        }
        else if ( t & ompt_task_final )
        {
            return "explicit_final";
        }
        else if ( t & ompt_task_mergeable )
        {
            return "explicit_mergeable";
        }
        else if ( t & ompt_task_merged )
        {
            return "explicit_merged";
        }
        else
        {
            assert( t == ompt_task_explicit );
            return "explicit";
        }
    }
    else if ( t & ompt_task_target )
    {
        if ( t & ompt_task_undeferred )
        {
            return "target_undeferred";
        }
        else if ( t & ompt_task_untied )
        {
            return "target_untied";
        }
        else if ( t & ompt_task_final )
        {
            return "target_final";
        }
        else if ( t & ompt_task_mergeable )
        {
            return "target_mergeable";
        }
        else if ( t & ompt_task_merged )
        {
            return "target_merged";
        }
        else
        {
            assert( t == ompt_task_target );
            return "target";
        }
    }
#if HAVE_DECL_OMPT_TASK_TASKWAIT
    else if ( t & ompt_task_taskwait )
    {
        if ( t & ompt_task_undeferred )
        {
            return "taskwait_undeferred";
        }
        else if ( t & ompt_task_untied )
        {
            return "taskwait_untied";
        }
        else if ( t & ompt_task_final )
        {
            return "taskwait_final";
        }
        else if ( t & ompt_task_mergeable )
        {
            return "taskwait_mergeable";
        }
        else if ( t & ompt_task_merged )
        {
            return "taskwait_merged";
        }
        else
        {
            assert( t == ompt_task_taskwait );
            return "taskwait";
        }
    }
#endif /* HAVE_DECL_OMPT_TASK_TASKWAIT */
    else
    {
        assert( false );
    }
    return "";
}

static const char*
task_status2string( ompt_task_status_t t )
{
    switch ( t )
    {
        case ompt_task_complete:
            return "complete";
        case ompt_task_yield:
            return "yield";
        case ompt_task_cancel:
            return "cancel";
        case ompt_task_detach:
            return "detach";
        case ompt_task_early_fulfill:
            return "early_fulfill";
        case ompt_task_late_fulfill:
            return "late_fulfill";
        case ompt_task_switch:
            return "switch";
        case ompt_taskwait_complete:
            return "taskwait_complete";
    }
    assert( false );
    return "";
}

static const char*
thread2string( ompt_thread_t t )
{
    switch ( t )
    {
        case ompt_thread_initial:
            return "initial";
        case ompt_thread_worker:
            return "worker";
        case ompt_thread_other:
            return "other";
        case ompt_thread_unknown:
            return "unknown";
    }
    assert( false );
    return "";
}

static const char*
work2string( ompt_work_t t )
{
    switch ( t )
    {
        case ompt_work_loop:
            return "loop";
        case ompt_work_sections:
            return "sections";
        case ompt_work_single_executor:
            return "single_executor";
        case ompt_work_single_other:
            return "single_other";
        case ompt_work_workshare:
            return "workshare";
        case ompt_work_distribute:
            return "distribute";
        case ompt_work_taskloop:
            return "taskloop";
        case ompt_work_scope:
            return "scope";
        #if HAVE_DECL_OMPT_WORK_LOOP_STATIC
        case ompt_work_loop_static:
            return "loop_static";
        #endif
        #if HAVE_DECL_OMPT_WORK_LOOP_DYNAMIC
        case ompt_work_loop_dynamic:
            return "loop_dynamic";
        #endif
        #if HAVE_DECL_OMPT_WORK_LOOP_GUIDED
        case ompt_work_loop_guided:
            return "loop_guided";
        #endif
        #if HAVE_DECL_OMPT_WORK_LOOP_OTHER
        case ompt_work_loop_other:
            return "loop_other";
        #endif
    }
    assert( false );
    return "";
}


// 5.2: callback interface entry points as in table 19.1
#define OMPT_CALLBACK_INTERFACE_ENTRY_POINTS \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( enumerate_states ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( enumerate_mutex_impls ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( set_callback ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_callback ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_thread_data ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_num_places ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_place_proc_ids ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_place_num ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_partition_place_nums ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_proc_id ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_state ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_parallel_info ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_task_info ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_task_memory ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_num_devices ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_num_procs ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_target_info ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( get_unique_id ) \
    OMPT_CALLBACK_INTERFACE_ENTRY_POINT( finalize_tool )

#define OMPT_CALLBACK_INTERFACE_ENTRY_POINT( NAME ) static ompt_ ## NAME ## _t NAME;
OMPT_CALLBACK_INTERFACE_ENTRY_POINTS
#undef OMPT_CALLBACK_INTERFACE_ENTRY_POINT

void
register_entry_points_callback_interface( ompt_function_lookup_t lookup )
{
    // 5.2, p538:2: In a compliant implementation, the lookup function
    // provided by the tool initializer for the OMPT callback
    // interface returns a valid function pointer for any OMPT runtime
    // entry point name listed in Table 19.1.

    #define OMPT_CALLBACK_INTERFACE_ENTRY_POINT( NAME ) NAME = ( ompt_ ## NAME ## _t )lookup( "ompt_" #NAME );
    OMPT_CALLBACK_INTERFACE_ENTRY_POINTS
    #undef OMPT_CALLBACK_INTERFACE_ENTRY_POINT

    #define OMPT_CALLBACK_INTERFACE_ENTRY_POINT( NAME ) \
    if ( NAME == NULL ) { \
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | callback interface entry point \"ompt_%s\" not provided\n", \
                                  __FUNCTION__, ompt_tool_tid, #NAME );       \
    }
    OMPT_CALLBACK_INTERFACE_ENTRY_POINTS
    #undef OMPT_CALLBACK_INTERFACE_ENTRY_POINT
}

/* Table 19.2: ompt_set_callback Must Return ompt_set_always */
/* Table 19.3: ompt_set_callback May Return Any Non-Error Code
   master is deprecated */
/* enum-value, name, signature, expected-result */
#define OMPT_CALLBACKS \
    OMPT_CALLBACK( 1, thread_begin, thread_begin, ompt_set_always )     \
    OMPT_CALLBACK( 2, thread_end, thread_end, ompt_set_always )         \
    OMPT_CALLBACK( 3, parallel_begin, parallel_begin, ompt_set_always ) \
    OMPT_CALLBACK( 4, parallel_end, parallel_end, ompt_set_always )     \
    OMPT_CALLBACK( 5, task_create, task_create, ompt_set_always )       \
    OMPT_CALLBACK( 6, task_schedule, task_schedule, ompt_set_always )   \
    OMPT_CALLBACK( 7, implicit_task, implicit_task, ompt_set_always )   \
    OMPT_CALLBACK( 8, target, target, ompt_set_always )                 \
    OMPT_CALLBACK( 33, target_emi, target_emi, ompt_set_always )                \
    OMPT_CALLBACK( 9, target_data_op, target_data_op, ompt_set_always ) \
    OMPT_CALLBACK( 33, target_data_op_emi, target_data_op_emi, ompt_set_always ) \
    OMPT_CALLBACK( 10, target_submit, target_submit, ompt_set_always )  \
    OMPT_CALLBACK( 35, target_submit_emi, target_submit_emi, ompt_set_always ) \
    OMPT_CALLBACK( 11, control_tool, control_tool, ompt_set_always )    \
    OMPT_CALLBACK( 12, device_initialize, device_initialize, ompt_set_always ) \
    OMPT_CALLBACK( 13, device_finalize, device_finalize, ompt_set_always ) \
    OMPT_CALLBACK( 14, device_load, device_load, ompt_set_always )      \
    OMPT_CALLBACK( 15, device_unload, device_unload, ompt_set_always )  \
    \
    OMPT_CALLBACK( 16, sync_region_wait, sync_region, ompt_set_error )  \
    OMPT_CALLBACK( 17, mutex_released, mutex, ompt_set_error )          \
    OMPT_CALLBACK( 18, dependences, dependences, ompt_set_error )       \
    OMPT_CALLBACK( 19, task_dependence, task_dependence, ompt_set_error ) \
    OMPT_CALLBACK( 20, work, work, ompt_set_error )                     \
    OMPT_CALLBACK( 21, masked, masked, ompt_set_error )                 \
    OMPT_CALLBACK( 22, target_map, target_map, ompt_set_error )         \
    OMPT_CALLBACK( 36, target_map_emi, target_map_emi, ompt_set_error ) \
    OMPT_CALLBACK( 23, sync_region, sync_region, ompt_set_error )       \
    OMPT_CALLBACK( 31, reduction, sync_region, ompt_set_error )         \
    OMPT_CALLBACK( 24, lock_init, mutex_acquire, ompt_set_error )       \
    OMPT_CALLBACK( 25, lock_destroy, mutex, ompt_set_error )            \
    OMPT_CALLBACK( 26, mutex_acquire, mutex_acquire, ompt_set_error )   \
    OMPT_CALLBACK( 27, mutex_acquired, mutex, ompt_set_error )          \
    OMPT_CALLBACK( 28, nest_lock, nest_lock, ompt_set_error )           \
    OMPT_CALLBACK( 29, flush, flush, ompt_set_error )                   \
    OMPT_CALLBACK( 30, cancel, cancel, ompt_set_error )                 \
    OMPT_CALLBACK( 32, dispatch, dispatch, ompt_set_error )             \
    OMPT_CALLBACK( 37, error, error, ompt_set_error )

void
thread_begin_cb( ompt_thread_t thread_type,
                 ompt_data_t*  thread_data )
{
    assert( ompt_tool_tid == -1 );
    static atomic_int_least32_t thread_counter = 0;
    ompt_tool_tid      = atomic_fetch_add( &thread_counter, 1 );
    thread_data->value = ompt_tool_tid;

    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | type = %s\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              thread2string( thread_type ) );
}

void
thread_end_cb( ompt_data_t* thread_data )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRIu32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
    assert( thread_data->value == ompt_tool_tid );
}

#define INVALID_TASK     6666666
#define INVALID_PARALLEL 7777777
#define INVALID_THREAD   8888888
#define INVALID_TARGET   9999999

void
parallel_begin_cb( ompt_data_t*        encountering_task_data,
                   const ompt_frame_t* encountering_task_frame,
                   ompt_data_t*        parallel_data,
                   unsigned int        requested_parallelism,
                   int                 flags,
                   const void*         codeptr_ra )
{
    static atomic_uint_least64_t parallel_counter = 7770001;
    parallel_data->value = atomic_fetch_add( &parallel_counter, 1 );

    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | encountering_task_data = %" PRIu64 " | flags = %s | requested_parallelism = %u | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data->value,
                              encountering_task_data == NULL ? INVALID_TASK : encountering_task_data->value,
                              parallel_flag2string( flags ),
                              requested_parallelism,
                              codeptr_ra );
}

void
parallel_end_cb( ompt_data_t* parallel_data,
                 ompt_data_t* encountering_task_data,
                 int          flags,
                 const void*  codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | encountering_task = %" PRIu64 " | flags = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              encountering_task_data == NULL ? INVALID_TASK : encountering_task_data->value,
                              parallel_flag2string( flags ),
                              codeptr_ra );
}

static uint64_t
new_task( void )
{
    static atomic_uint_least64_t task_counter = 6660001;
    return atomic_fetch_add( &task_counter, 1 );
}

void
task_create_cb( ompt_data_t*        encountering_task_data,
                const ompt_frame_t* encountering_task_frame,
                ompt_data_t*        new_task_data,
                int                 flags,
                int                 has_dependences,
                const void*         codeptr_ra )
{
    assert( new_task_data->value == 0 );
    new_task_data->value = new_task();

    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | encountering_task_data = %" PRIu64 " | new_task_data = %" PRIu64 " | flags = %s | has_dependences = %d | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              encountering_task_data->value,
                              new_task_data->value,
                              task_flag2string( flags ),
                              has_dependences,
                              codeptr_ra );
}

void
task_schedule_cb( ompt_data_t*       prior_task_data,
                  ompt_task_status_t prior_task_status,
                  ompt_data_t*       next_task_data )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | prior_task_data = %" PRIu64 " | prior_status = %s | next_task_data = %" PRIu64 "\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              prior_task_data->value,
                              task_status2string( prior_task_status ),
                              next_task_data->value );
}

void
implicit_task_cb( ompt_scope_endpoint_t endpoint,
                  ompt_data_t*          parallel_data,
                  ompt_data_t*          task_data,
                  unsigned int          actual_parallelism,
                  unsigned int          index, /* For initial tasks, that are not created
                                                  by a teams construct, this argument is 1. */
                  int                   flags )
{
    if ( endpoint == ompt_scope_begin )
    {
        uint64_t old = task_data->value;
        task_data->value = new_task();
        if( old != 0 )
        {
            // runtime might reuse implicit tasks but doesn't reset
            // ompt_data_t. Seems to be legal. Seem with NVHPC 23.1.
            OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " (reused: %" PRIu64 ") | endpoint = %s | actual_parallelism = %u | index = %u | flags = %s\n",
                                      __FUNCTION__,
                                      ompt_tool_tid,
                                      parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                                      task_data == NULL ? INVALID_TASK : task_data->value,
                                      old,
                                      scope_endpoint2string( endpoint ),
                                      actual_parallelism,
                                      index,
                                      task_flag2string( flags ) );

            return;
        }


    }
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | endpoint = %s | actual_parallelism = %u | index = %u | flags = %s\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              scope_endpoint2string( endpoint ),
                              actual_parallelism,
                              index,
                              task_flag2string( flags ) );
}

void
target_cb( ompt_target_t         kind,
           ompt_scope_endpoint_t endpoint,
           int                   device_num,
           ompt_data_t*          task_data,
           ompt_id_t             target_id,
           const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | endpoint = %s | kind = %s | device_num = %d | target_id = %" PRIu64 " | task_data = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              scope_endpoint2string( endpoint ),
                              target2string( kind ),
                              device_num,
                              target_id,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              codeptr_ra );
}

void
target_emi_cb( ompt_target_t         kind,
               ompt_scope_endpoint_t endpoint,
               int                   device_num,
               ompt_data_t*          task_data,
               ompt_data_t*          target_task_data,
               ompt_data_t*          target_data,
               const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | endpoint = %s | kind = %s | device_num = %d | task_data = %" PRIu64 " | target_task_data = %" PRIu64 " | target_data = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              scope_endpoint2string( endpoint ),
                              target2string( kind ),
                              device_num,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              target_task_data == NULL ? INVALID_TASK : target_task_data->value,
                              target_data == NULL ? INVALID_TARGET : target_data->value,
                              codeptr_ra );
}

void
target_data_op_cb( ompt_id_t             target_id,
                   ompt_id_t             host_op_id,
                   ompt_target_data_op_t optype,
                   void*                 src_addr,
                   int                   src_device_num,
                   void*                 dest_addr,
                   int                   dest_device_num,
                   size_t                bytes,
                   const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | target_id = %" PRIu64 " | host_op_id = %"  PRIu64 " | src_addr = %p | dest_addr = %p | src_device_num = %d | dest_device_num = %d | optype = %s | bytes = %lu | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              target_id,
                              host_op_id,
                              src_addr,
                              dest_addr,
                              src_device_num,
                              dest_device_num,
                              target_data_op2string( optype ),
                              bytes,
                              codeptr_ra );
}

void
target_data_op_emi_cb( ompt_scope_endpoint_t endpoint,
                       ompt_data_t*          target_task_data,
                       ompt_data_t*          target_data,
                       ompt_id_t*            host_op_id,
                       ompt_target_data_op_t optype,
                       void*                 src_addr,
                       int                   src_device_num,
                       void*                 dest_addr,
                       int                   dest_device_num,
                       size_t                bytes,
                       const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | endpoint = %s | target_task_data = %" PRIu64 " | target_data = %" PRIu64 " | host_op_id = %" PRIu64 " | src_addr = %p | dest_addr = %p | src_device_num = %d | dest_device_num = %d | optype = %s | bytes = %lu | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              scope_endpoint2string( endpoint ),
                              target_task_data == NULL ? INVALID_TASK : target_task_data->value,
                              target_data == NULL ? INVALID_TARGET : target_data->value,
                              *host_op_id,
                              src_addr,
                              dest_addr,
                              src_device_num,
                              dest_device_num,
                              target_data_op2string( optype ),
                              bytes,
                              codeptr_ra );
}

void
target_submit_cb( ompt_id_t    target_id,
                  ompt_id_t    host_op_id,
                  unsigned int requested_num_teams )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
}

void
target_submit_emi_cb( ompt_scope_endpoint_t endpoint,
                      ompt_data_t*          target_data,
                      ompt_id_t*            host_op_id,
                      unsigned int          requested_num_teams )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | endpoint = %s\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              scope_endpoint2string( endpoint ) );
}

void
control_tool_cb( uint64_t    command,
                 uint64_t    modifier,
                 void*       arg,
                 const void* codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              codeptr_ra );
}


// 5.2. Table 19.4: per device runtime entry points
#define OMPT_TRACING_INTERFACE_ENTRY_POINTS \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( get_device_num_procs )  \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( get_device_time ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( translate_time ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( set_trace_ompt ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( set_trace_native ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( start_trace ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( pause_trace ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( flush_trace ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( stop_trace ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( advance_buffer_cursor ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( get_record_type ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( get_record_ompt ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( get_record_native ) \
    OMPT_TRACING_INTERFACE_ENTRY_POINT( get_record_abstract )


typedef struct device_t device_t;
struct device_t
{
    int            device_num;
    const char*    type;
    ompt_device_t* device;
    const char*    documentation;
    ompt_buffer_t* buffer;
    device_t*      next;

    #define OMPT_TRACING_INTERFACE_ENTRY_POINT( NAME ) ompt_ ## NAME ## _t NAME;
    OMPT_TRACING_INTERFACE_ENTRY_POINTS
    #undef OMPT_TRACING_INTERFACE_ENTRY_POINT
};

/* singly linked list of initialized devices */
static device_t*   devices;
static atomic_flag devices_mutex = ATOMIC_FLAG_INIT;

static void
device_add( device_t* d )
{
    OMPT_TOOL_LOCK( devices_mutex );
    d->next = devices;
    devices = d;
    OMPT_TOOL_UNLOCK( devices_mutex );
}

static device_t*
device_remove( int device_num )
{
    OMPT_TOOL_LOCK( devices_mutex );
    if ( !devices )
    {
        OMPT_TOOL_UNLOCK( devices_mutex );
        return NULL;
    }
    if ( devices->device_num == device_num )
    {
        device_t* curr = devices;
        devices = devices->next;
        OMPT_TOOL_UNLOCK( devices_mutex );
        return curr;
    }
    device_t* curr = devices;
    device_t* prev = NULL;
    while ( curr != NULL && curr->device_num != device_num )
    {
        prev = curr;
        curr = curr->next;
    }
    if ( curr == NULL )
    {
        OMPT_TOOL_UNLOCK( devices_mutex );
        return NULL;
    }
    prev->next = curr->next;
    OMPT_TOOL_UNLOCK( devices_mutex );
    return curr;
}


#define TRACE_BUF_SIZE ( 1024 * 1024 )

static void
callback_buffer_request( int             device_num,
                         ompt_buffer_t** buffer,
                         size_t*         bytes )
{
    /* If the device requires a trace buffer, the OpenMP
       implementation invokes the tool-supplied callback function on
       the host device to request a new buffer. */
    /* It is possible to share a buffer between devices (how? why?). */

    *bytes  = TRACE_BUF_SIZE;
    *buffer = malloc( *bytes );
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "| trace buffer for device %d allocated\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              device_num );
}

static void
callback_buffer_complete( int                  device_num,
                          ompt_buffer_t*       buffer,
                          size_t               bytes,
                          ompt_buffer_cursor_t begin,
                          int                  buffer_owned )
{
    /* When appropriate, for example, when a trace buffer fills or
       needs to be flushed, the OpenMP implementation invokes the
       tool-supplied buffer completion callback to process a non-empty
       sequence of records in a trace buffer that is associated with
       the device. */

    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );

    // TODO
    /* iterate through buffer using the ompt_advance_buffer_cursor entry point */
    /* query record type (ompt, native, invalid) using the ompt_get_record_type runtime entry point */
    /* inspect an OMPT record using the ompt_get_record_ompt runtime entry point */

    /* The 5.2 specification does not mention automatic buffer reuse */
    if ( buffer_owned == 1 )
    {
        free( buffer );
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | trace buffer of device %d freed\n",
                                  __FUNCTION__,
                                  ompt_tool_tid,
                                  device_num );
    }
    else
    {
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | trace buffer of device %d can't be freed\n",
                                  __FUNCTION__,
                                  ompt_tool_tid,
                                  device_num );
    }
}

void
device_initialize_cb( int                    device_num,
                      const char*            type,
                      ompt_device_t*         device,
                      ompt_function_lookup_t lookup, /* device specific */
                      const char*            documentation )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
    // if lookup == NULL, device does not support tracing, otherwise
    // lookup tracing interface entry points (per device) as in Table
    // 19.4
    if ( lookup )
    {
        device_t* d = malloc( sizeof( *d ) );
        d->device_num    = device_num;
        d->type          = type;
        d->device        = device;
        d->documentation = documentation;
        d->buffer        = NULL;
        d->next          = NULL;

        #define OMPT_TRACING_INTERFACE_ENTRY_POINT( NAME ) d->NAME = ( ompt_ ## NAME ## _t )lookup( "ompt_" #NAME );
        OMPT_TRACING_INTERFACE_ENTRY_POINTS
        #undef OMPT_TRACING_INTERFACE_ENTRY_POINT

        #define OMPT_TRACING_INTERFACE_ENTRY_POINT( NAME ) \
    if ( d->NAME == NULL ) \
    { \
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | tracing interface entry point \"ompt_%s\" not provided for %s(%d)\n", \
                                  __FUNCTION__, ompt_tool_tid, #NAME, type, device_num ); \
    }
        OMPT_TRACING_INTERFACE_ENTRY_POINTS
        #undef OMPT_TRACING_INTERFACE_ENTRY_POINT

        device_add( d );

        // ompt_set_trace_ompt and ompt_get_record_ompt should be
        // available to trace in a portable format.
        assert( d->set_trace_ompt );
        assert( d->get_record_ompt );
        /* Try to enable device records:
           Enables Each OMPT trace record serves as a substitute for an OMPT
           callback that cannot be made on the device. */
        ompt_set_result_t result;
        #define OMPT_CALLBACK( ID, NAME, SIGNATURE, EXPECTED ) \
    result = d->set_trace_ompt( device, 1 /* enable */, ID ); \
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | enabling trace record %16s for %s(%d) results in \"%s\".\n", \
                              __FUNCTION__, ompt_tool_tid, #NAME, type, device_num, set_result2string( result ) );
        OMPT_CALLBACKS
        #undef OMPT_CALLBACK

        // start tracing by invoking ompt_start_trace (at any time)
        if ( d->start_trace )
        {
            int started = d->start_trace( d->device,
                                          callback_buffer_request,
                                          callback_buffer_complete );
            assert( started );
        }
    }
    else
    {
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | %s(%d) does not provide a lookup function, i.e. OMPT tracing not supported.\n",
                                  __FUNCTION__,
                                  ompt_tool_tid,
                                  type,
                                  device_num );
    }
}

void
device_finalize_cb( int device_num )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
    device_t* d = device_remove( device_num );
    if ( d )
    {
        free( d );
    }
}

void
device_load_cb( int         device_num,
                const char* filename,
                int64_t     offset_in_file,
                void*       vma_in_file,
                size_t      bytes,
                void*       host_addr,
                void*       device_addr,
                uint64_t    module_id )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
}

void
device_unload_cb( int      device_num,
                  uint64_t module_id )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
}

void
sync_region_wait_cb( ompt_sync_region_t    kind,
                     ompt_scope_endpoint_t endpoint,
                     ompt_data_t*          parallel_data,
                     ompt_data_t*          task_data,
                     const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | endpoint = %s | kind = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              scope_endpoint2string( endpoint ),
                              sync_region2string( kind ),
                              codeptr_ra );
}

void
mutex_released_cb( ompt_mutex_t   kind,
                   ompt_wait_id_t wait_id,
                   const void*    codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | kind = %s | wait_id = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              mutex2string( kind ),
                              wait_id,
                              codeptr_ra );
}

void
dependences_cb( ompt_data_t*             task_data,
                const ompt_dependence_t* deps,
                int                      ndeps )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | ndpes = %d\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              ndeps );
}

void
task_dependence_cb( ompt_data_t* src_task_data,
                    ompt_data_t* sink_task_data )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | src_task_data = %" PRIu64 " | sink_task_data = %" PRIu64 "\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              src_task_data == NULL ? INVALID_TASK : src_task_data->value,
                              sink_task_data == NULL ? INVALID_TASK : sink_task_data->value );
}

void
work_cb( ompt_work_t           work_type,
         ompt_scope_endpoint_t endpoint,
         ompt_data_t*          parallel_data,
         ompt_data_t*          task_data,
         uint64_t              count,
         const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | endpoint = %s | work_type = %s | count = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              scope_endpoint2string( endpoint ),
                              work2string( work_type ),
                              count,
                              codeptr_ra );
}

void
masked_cb( ompt_scope_endpoint_t endpoint,
           ompt_data_t*          parallel_data,
           ompt_data_t*          task_data,
           const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | endpoint = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              scope_endpoint2string( endpoint ),
                              codeptr_ra );
}

void
target_map_cb( ompt_id_t     target_id,
               unsigned int  nitems,
               void**        host_addr,
               void**        device_addr,
               size_t*       bytes,
               unsigned int* mapping_flags,
               const void*   codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | mapping_flags = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              target_map_flag2string( *mapping_flags ),
                              codeptr_ra );
}

void
target_map_emi_cb( ompt_data_t*  target_data,
                   unsigned int  nitems,
                   void**        host_addr,
                   void**        device_addr,
                   size_t*       bytes,
                   unsigned int* mapping_flags,
                   const void*   codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | mapping_flags = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              target_map_flag2string( *mapping_flags ),
                              codeptr_ra );
}

void
sync_region_cb( ompt_sync_region_t    kind,
                ompt_scope_endpoint_t endpoint,
                ompt_data_t*          parallel_data,
                ompt_data_t*          task_data,
                const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | endpoint = %s | kind = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              scope_endpoint2string( endpoint ),
                              sync_region2string( kind ),
                              codeptr_ra );
}

void
reduction_cb( ompt_sync_region_t    kind,
              ompt_scope_endpoint_t endpoint,
              ompt_data_t*          parallel_data,
              ompt_data_t*          task_data,
              const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | endpoint = %s | kind = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              scope_endpoint2string( endpoint ),
                              sync_region2string( kind ),
                              codeptr_ra );
}

void
lock_init_cb( ompt_mutex_t   kind,
              unsigned int   hint,
              unsigned int   impl,
              ompt_wait_id_t wait_id,
              const void*    codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | kind = %s | wait_id = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              mutex2string( kind ),
                              wait_id,
                              codeptr_ra );
}

void
lock_destroy_cb( ompt_mutex_t   kind,
                 ompt_wait_id_t wait_id,
                 const void*    codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | kind = %s | wait_id = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              mutex2string( kind ),
                              wait_id,
                              codeptr_ra );
}

void
mutex_acquire_cb( ompt_mutex_t   kind,
                  unsigned int   hint,
                  unsigned int   impl,
                  ompt_wait_id_t wait_id,
                  const void*    codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | kind = %s | wait_id = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              mutex2string( kind ),
                              wait_id,
                              codeptr_ra );
}

void
mutex_acquired_cb( ompt_mutex_t   kind,
                   ompt_wait_id_t wait_id,
                   const void*    codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | kind = %s | wait_id = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              mutex2string( kind ),
                              wait_id,
                              codeptr_ra );
}

void
nest_lock_cb( ompt_scope_endpoint_t endpoint,
              ompt_wait_id_t        wait_id,
              const void*           codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | endpoint = %s | wait_id = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              scope_endpoint2string( endpoint ),
                              wait_id,
                              codeptr_ra );
}

void
flush_cb( ompt_data_t* thread_data,
          const void*  codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | thread_data = %" PRIu64 " | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              thread_data == NULL ? INVALID_THREAD : thread_data->value,
                              codeptr_ra );
}

void
cancel_cb( ompt_data_t* task_data,
           int          flags,
           const void*  codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | task_data = %" PRIu64 " | flags = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              cancel_flag2string( flags ),
                              codeptr_ra );
}

void
dispatch_cb( ompt_data_t*    parallel_data,
             ompt_data_t*    task_data,
             ompt_dispatch_t kind,
             ompt_data_t     instance )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | parallel_data = %" PRIu64 " | task_data = %" PRIu64 " | kind = %s\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              parallel_data == NULL ? INVALID_PARALLEL : parallel_data->value,
                              task_data == NULL ? INVALID_TASK : task_data->value,
                              dispatch2string( kind ) );
}

#if HAVE_OMPT_SEVERITY_T
void
error_cb( ompt_severity_t severity,
          const char*     message,
          size_t          length,
          const void*     codeptr_ra )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | severity = %s | message = %s | codeptr_ra = %p\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              severity2string( severity ),
                              message,
                              codeptr_ra );
}
#endif

void
register_callbacks( void )
{
    assert( set_callback );
    ompt_set_result_t result;
    #define OMPT_CALLBACK( ID, NAME, SIGNATURE, EXPECTED ) \
    result = set_callback( ompt_callback_ ## NAME, ( ompt_callback_t )&NAME ## _cb ); \
    if ( EXPECTED == ompt_set_always ) \
    { \
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | %18s = %s (expect always)\n", \
                                  __FUNCTION__, ompt_tool_tid, #NAME, set_result2string( result ) ); \
    } \
    else \
    { \
        OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | %18s = %s\n", \
                                  __FUNCTION__, ompt_tool_tid, #NAME, set_result2string( result ) ); \
    }
    OMPT_CALLBACKS
    #undef OMPT_CALLBACK
}


static int
my_initialize_tool( ompt_function_lookup_t lookup,
                    int                    initial_device_num,
                    ompt_data_t*           tool_data )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | initial_device_num %d\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              initial_device_num );
    register_entry_points_callback_interface( lookup );
    register_callbacks();
    /* tracing runtime entry points are registered per device via device_initialize_cb */

    return 1; /* non-zero indicates success */
}

void
ompt_tool_trigger_ompt_finalize_tool( void )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
    finalize_tool();
}

static void
my_finalize_tool( ompt_data_t* tool_data )
{
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 "\n",
                              __FUNCTION__,
                              ompt_tool_tid );
}

ompt_start_tool_result_t*
ompt_start_tool( unsigned int omp_version,
                 const char*  runtime_version )
{
    setbuf( stdout, NULL );
    OMPT_TOOL_GUARDED_PRINTF( "[%s] tid = %" PRId32 " | omp_version %d | runtime_version = \'%s\'\n",
                              __FUNCTION__,
                              ompt_tool_tid,
                              omp_version,
                              runtime_version );
    static ompt_start_tool_result_t tool = { &my_initialize_tool,
                                             &my_finalize_tool,
                                             ompt_data_none };
    return &tool;
}
